----------------------------------------------------------------------------------------------------
-- RunState
----------------------------------------------------------------------------------------------------
-- Purpose: Run state for the state machine.
--
-- Copyright (c) Wopster, 2020
----------------------------------------------------------------------------------------------------

---@class RunState
---@field public id number
RunState = {}
RunState.PENALTY_TIME = 5000 --ms

local RunState_mt = Class(RunState, AbstractState)

---Creates a new run state.
---@param id number
---@param stateMachine FSM
---@param custom_mt table
---@return RunState
function RunState:new(id, stateMachine, isServer, coneHandler, trafficLight, checkpoints, hud, vehicleList, object, custom_mt)
    local self = AbstractState:new(id, stateMachine, custom_mt or RunState_mt)

    -- Create action here to be detected by the cone handler.
    self.action = function(count)
        self:onPenaltyCountChanged(count)
    end

    self.isServer = isServer
    self.coneHandler = coneHandler
    self.trafficLight = trafficLight

    self.checkpoints = checkpoints
    self.hud = hud
    self.vehicleList = vehicleList
    self.object = object
    self.playerUserId = nil

    self.queue = Queue:new()
    self.timer = Timer:new()
    self.timer:setIsStopWatch(true)

    return self
end

---Called on delete.
function RunState:delete()
    RunState:superClass().delete(self)
end

---Called on enter.
function RunState:onEnter(playerUserId)
    self.playerUserId = playerUserId

    RunState:superClass().onEnter(self)
    self.trafficLight:next() -- to GREEN.

    if self.isServer then
        for _, vehicle in ipairs(self.vehicleList) do
            if vehicle.setAllowDriveSuperCup ~= nil then
                vehicle:setAllowDriveSuperCup(true, self.object)
            end
        end
    end

    self.coneHandler:addActionListener(self.action)

    self.hud:setIsActive(true)
    self.hud:setTimer(self.timer)
    self.hud:setPenaltyPoints(0)

    self.timer:start()

    if self.isServer then
        self.queue = Queue.fromList(ListUtil.copyTable(self.checkpoints))
        self.currentCheckpoint = self.queue:pop()
        self.currentCheckpoint:setIsActive(true)
    end
end

---Called on exit.
function RunState:onExit()
    RunState:superClass().onExit(self)
    self.coneHandler:removeActionListener(self.action)
    self.timer:stop()
end

function RunState:onUpdate(dt)
    self.coneHandler:update(dt)

    if self.isServer then
        if self.currentCheckpoint ~= nil then
            if self.currentCheckpoint:hasPassedCheckpoint() then
                self.currentCheckpoint:setIsActive(false)
                self.currentCheckpoint = self.queue:pop()

                if self.currentCheckpoint ~= nil then
                    self.currentCheckpoint:setIsActive(true)
                end
            end
        else
            local isStopped = 0 -- false
            self.stateMachine:transition(SuperCup.STATES.FINISHED, { self.playerUserId, self.timer:getElapsedTime(), self.coneHandler:getLostBallCount() })
        end
    end
end

function RunState:onPenaltyCountChanged(count)
    self.timer.elapsedDelay = self.timer.elapsedDelay + RunState.PENALTY_TIME
    self.hud:setPenaltyPoints(count)
end
